/******************************************************************************
 * AudioTask.c
 * Copyright (C) Infineon Technologies AG 2021
 *
 * Use of this file is subject to the terms of use agreed between (i) you or
 * the company in which ordinary course of business you are acting and (ii)
 * Infineon Technologies AG or its licensees. If and as long as no such terms
 * of use are agreed, use of this file is subject to following:
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated
 * by a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *****************************************************************************/


/*****************************************************************************/
/*----------------------------------Includes---------------------------------*/
/*****************************************************************************/
#pragma optimize 0

#include <stdint.h>
#include <math.h>
#include "Audio.h"
#include "ExtControl.h"
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include <Ifx_reg.h>
#include "Ifx_Types.h"
#include "OpenPDMFilter.h"
#include "model.h"
#include <string.h>
#include "UartTask.h"
#include <stdio.h>
#include "IfxStm.h"


/* lwIP includes. */
#include "lwip/api.h"
#include "lwip/tcpip.h"
#include "lwip/memp.h"
#include "lwip/stats.h"

#include "lwip/apps/lwiperf.h"

#include "lwip/tcp.h"
#include "lwip/udp.h"
#include "lwip/sys.h"

/*****************************************************************************/
/*-----------------------------------Macros----------------------------------*/
/*****************************************************************************/

/*****************************************************************************/
/*------------------------------Global variables-----------------------------*/
/*****************************************************************************/

TaskHandle_t AudioProcess_task_handler = NULL;

//Audio RX/TX Buffer
extern int32_t __attribute__ ((__align(4*128))) rxBuf[128];
extern int32_t __attribute__ ((__align(4*128))) txBuf[128];

//PDM Buffers
uint16_t t1, t2;
uint16_t rxBufSplit[2][64];
int16_t MICPCMBuffer[2][16];


//Volume
float volume=1.0f;

//decide here if the poti/adc should be used for volume-control
uint8_t useADCVolumeControl = 0;

/*LwIP IP and Port*/
static struct udp_pcb * new_udp_pcb;
static unsigned short port, dest_port;
static ip_addr_t  addr, dest_addr;

/*****************************************************************************/
/*-------------------------------Data Structures-----------------------------*/
/*****************************************************************************/

/*****************************************************************************/
/*-------------------------Private Variables/Constants-----------------------*/
/*****************************************************************************/

/*****************************************************************************/
/*----------------------------Function Prototypes----------------------------*/
/*****************************************************************************/

/*****************************************************************************/
/*-------------------------Function Implementations--------------------------*/
/*****************************************************************************/



//PDM Filter struct
TPDMFilter_InitStruct filter0;


int16_t SampleSaturate (float inSample) {

    /*
     * Max sample value can be +/- 2^15. So before back-conversion from float to int_16 we need to saturate the sample (otherwise distorted sound might appear)
     */

    if (inSample > 32700.0f) inSample = 32700.0f;
    if (inSample < -32700.0f) inSample = -32700.0f;
    return (int16_t) roundf(inSample);
}


IFX_INLINE void bmerge (uint16_t in_odd, uint16_t in_even, uint32_t * out_data) {

    /*
     * This function is used to double up the PCM bits as the QSPI is running at double-speed.
     * Please check the user's manual for more information.
     */

    uint32_t merged_value;

    __asm(
            "bmerge %0, %1, %2 \n\t"
            : "=d" (merged_value)
            : "d" (in_odd), "d" (in_even)
          );

    *out_data = merged_value;
}


IFX_INLINE void bsplit(uint32_t indata, uint16_t * out_even, uint16_t * out_odd) {

    /*
     * This function is used to extract the interleaved PDM-data stream into to different buffers
     * Please check the user's manual for more information.
     */
    __asm( "bsplit e14, %0 \n\t"
            "st.h [%1],d14\n\t"
            "st.h [%2],d15\n\t"
            :
            : "d" (indata), "a" (out_even), "a" (out_odd) : "e14"
          );

}

uint16_t bswap16 (uint16_t in) {

    /*
     * Swaps the hi- and low byte
     */

    uint16_t res =
            ((in & 0x00FF)<<8) |
            ((in & 0xFF00)>>8);
    return res;
}

/*
 * Interrupt from RX-DMA at every buffer-switchover (as of current configuration every 64 samples (or 32 stereo-samples)
 */
IFX_INTERRUPT (DMA_INTERRUPT, 0, 30)
{

    //Clear DMA Frozen Bit for I�S RX/TX DMA
    DMA_CHCSR050.B.FROZEN = 0;
    DMA_CHCSR049.B.FROZEN = 0;

    //Unblock Audio Process Task and force context switch
    BaseType_t xHigherPriorityTaskWoken = pdFALSE;
    vTaskNotifyGiveFromISR( AudioProcess_task_handler, &xHigherPriorityTaskWoken );
    portYIELD_FROM_ISR( xHigherPriorityTaskWoken );

}

#define DATA_PROCESS_QUEUE_SIZE 10000

float audioData [DATA_PROCESS_QUEUE_SIZE];
float classifications[IMAI_DATA_OUT_COUNT];

uint64 myTimingPups = 0;
uint64 prevStm = 0;
uint32 head     = 0;
float divider = 4096.0;
struct pbuf *p;
err_t error;

static __attribute__((__noreturn__)) void OsTask_AudioProcess(void *arg)
{
    (void)arg;
    uint64 curStm;
    extern err_t ethernetif_init( struct netif *netif );

    //Inits I2C
    I2CInit();

    //Use oscillator for fixed 48 kHz sample-rate
    InitRefClockMux(0);

    //Init PDM Filter
    filter0.Fs = 16000; // 48000;
    filter0.LP_HZ = 20000;
    filter0.HP_HZ = 10;
    filter0.In_MicChannels = 1;     //warning: library was modified to stereo-pdm operation
    filter0.Out_MicChannels = 1;    //warning: library was modified to stereo-pdm operation
    filter0.Decimation = 64;
    filter0.MaxVolume = 64;
    Open_PDM_Filter_Init(&filter0);

    //start I2S emulation; critical section should not be interrupted by context switch
    taskENTER_CRITICAL();
    I2S_Init_DMA();
    taskEXIT_CRITICAL();

    //init Amp and Codec
    InitAmpAndCodec();

    uint8_t ActiveBuffer;

    /*UDP Connection*/
    new_udp_pcb = udp_new();
    IP_ADDR4(&dest_addr, 192, 168, 2, 10);
    IP_ADDR4(&addr, 192, 168, 2, 2018);
    dest_port = 4444;
    port = 51841;
    udp_bind(new_udp_pcb, IP_ADDR_ANY, port);
    udp_connect(new_udp_pcb, &dest_addr, dest_port);


    while (1)
    {
        //wait for task unblock from ISR
        ulTaskNotifyTake(pdFALSE, portMAX_DELAY);

        //check DMA which buffer we should operate on
        if (DMA_CHCSR050.B.BUFFER == 0) ActiveBuffer = 1;
        else ActiveBuffer = 0;

        //calculate buffer offset dependent if DMA currently operates on first or second buffer half
        int BufferOffset = ActiveBuffer ? 64 : 0;


        //Get PDM Data from input-buffer, split them and apply the bit-swap
        int j = 0;
        for (int i=0+BufferOffset; i<64+BufferOffset; i++) {
            bsplit(rxBuf[i],&t1, &t2);
            rxBufSplit[0][j] = bswap16(t1);
            rxBufSplit[1][j] = bswap16(t2);
            j++;
        }

        //Filter the PDM data to PCM
        Open_PDM_Filter_64((uint8_t*) &rxBufSplit[0], (uint8_t*) &rxBufSplit[1], (int16_t*) &MICPCMBuffer[0], (int16_t*) &MICPCMBuffer[1], (uint16_t)50, (TPDMFilter_InitStruct *) &filter0);


        //Apply volume-modification
        for (int i=0; i<16; i++)
        {
            MICPCMBuffer[0][i] = SampleSaturate( ((float) MICPCMBuffer[0][i])*volume );
            MICPCMBuffer[1][i] = SampleSaturate( ((float) MICPCMBuffer[1][i])*volume );

            /*Saturate Data and adjust the volume*/
            float audio_sample = (float)MICPCMBuffer[0][i] / divider;
            /*push the PCM data into the circular buffer*/
            audioData[head] = audio_sample;
            head = (head+1) % DATA_PROCESS_QUEUE_SIZE;

        }

        /*Send the PCM data via the UDP connection as a raw stream*/
        p = pbuf_alloc(PBUF_TRANSPORT, 4096, PBUF_RAM);
        p->len = 32;
        p->tot_len = 32;
        memcpy(p->payload, MICPCMBuffer[0], 32);
        udp_send(new_udp_pcb,p);
        pbuf_free(p);

        //Double up the PCM bits as I�S interface data-output runs effectively at double-speed and write to I�S TX-Buffer
        uint32_t out_data;
        j=0+BufferOffset;
        for (int i=0; i<16;i++)
        {
            bmerge(MICPCMBuffer[0][i], MICPCMBuffer[0][i], &out_data);
            txBuf[j] = out_data;
            txBuf[j+1] = 0;
            bmerge(MICPCMBuffer[1][i], MICPCMBuffer[1][i], &out_data);
            txBuf[j+2] = out_data;
            txBuf[j+3] = 0;
            j+=4;
        }
    }
}


static __attribute__((__noreturn__)) void OsTask_ExternalAudioControl(void *arg)
{

    (void)arg;

    InitButtonADC();

    //Check if user-button pressed at startup -> if yes activate volume control over poti
    if (UserButtonPressedStartup() == 0) {
        useADCVolumeControl = 1;
    }
    else {
        useADCVolumeControl = 0;
        volume = 1.0f;
    }

    while (1) {

        //check if users wants to mute/unmute amplifier
        CheckAmpMute();

        //read volume from poti
        if (useADCVolumeControl==1) {
            volume = ReadADCVolume();
        }

        //Delay 100ms
        vTaskDelay(pdMS_TO_TICKS(100));
    }

}


uint32 realHead = 0;
uint32 tail     = 0;
uint32 distance = 0;

static __attribute__((__noreturn__)) void OsTask_DataProcessing(void *arg)
{
    (void)arg;

    /* Initialize the Machine Learning model */
    IMAI_init();

    while(1)
    {
        /*Handling of Circular Buffer*/
        do
        {
            taskENTER_CRITICAL();
            realHead = head >= tail ? head : head + DATA_PROCESS_QUEUE_SIZE;
            taskEXIT_CRITICAL();

            if (distance < (realHead - tail) )
            { distance = (realHead - tail); }

            if (tail >= realHead)
            {
                /*Delay 1ms*/
                vTaskDelay(pdMS_TO_TICKS(1));
            }
        } while (tail >= realHead);

        /*Get the PCM data from the circular buffer and adjust the volume*/
        float sample = audioData[tail] * 0.125f;
        /*enqueue the PCM data to Siren Detection Model*/
        if(IMAI_enqueue(&sample) != IMAI_RET_SUCCESS)
        {
            /* error occurred */
            while(1);
        }
        tail = (tail + 1) % DATA_PROCESS_QUEUE_SIZE;

        /*dequeue the PCM data from Siren Detection Model*/
        while(IMAI_dequeue(classifications) == IMAI_RET_SUCCESS)
        {
#if IMAI_DATA_OUT_COUNT == 2
            if(classifications[1] > 0.70)
            {
                /*print the recognition results with recognition >= 70%
                 *reduce the load for printing on the serial Interface*/
                sprintf(UART_taskBuf, "None:\t %.2f Siren:\t %.2f \n\r", classifications[0], classifications[1]);
            }
#elif IMAI_DATA_OUT_COUNT == 3
            if(classifications[2] > 0.75 || classifications[1] > 0.75)
            {
                sprintf(UART_taskBuf, "None:\t %.2f  Down:\t %.2f  Up:\t%.2f \n\r", classifications[0], classifications[1], classifications[2]);
            }
#endif
        }
    }
}

void AudioTasksCreate(void) {

    //Create Tasks
    xTaskCreate(OsTask_DataProcessing, "DataProcessing task", configMINIMAL_STACK_SIZE, NULL, configMAX_PRIORITIES - 3, NULL);
    xTaskCreate(OsTask_AudioProcess, "Audio task", configMINIMAL_STACK_SIZE, NULL, configMAX_PRIORITIES - 1, &AudioProcess_task_handler);
    xTaskCreate(OsTask_ExternalAudioControl, "Ext. Control task", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY, NULL);

}
